from matplotlib import pyplot as plt
import numpy as np
import os
import shutil
from tqdm import tqdm
import copy

from .aero import Aero
from .geometry import Plane, Section


class Dihedral():
    def __init__(self, dihedral_config_file: str, aero_config_file: str):

        #   Clean temp folders.
        self.path = os.path.split(dihedral_config_file)[0]

        try:
            if os.path.isdir(self.path+"/results") == True:
                shutil.rmtree(self.path+"/results")
            if os.path.isdir(self.path+"/generated planes") == True:
                shutil.rmtree(self.path+"/generated planes")
            if os.path.isdir(self.path+"/cases") == True:
                shutil.rmtree(self.path+"/cases")
        except PermissionError:
            raise PermissionError("Close all results, geometry, case files")

        os.mkdir(self.path+"/generated planes")
        os.mkdir(self.path+"/results")
        os.mkdir(self.path+"/cases")

        if os.path.exists(f"{self.path}/avl.exe") == False:
            print("\u001b[31m[Error]\u001b[0m avl.exe not found.")
            exit()

        self.read_config(dihedral_config_file)
        self.aero_config_file = aero_config_file

        return None

    def read_config(self, config_file):
        """
        Reads dihedral config file.

        Arguments:
            config_file: {string} -- Dihedral config file.
        """
        def str_to_bool(x): return True if (x == "Y") else False

        with open(config_file, 'r') as f:
            lines = f.readlines()
        lines = [line for line in lines if line[0]
                 != "#" and line != "\n"]  # cleans input

        if lines[0].strip() != "DIHEDRAL CONFIG":
            print(
                f"\u001b[31m[Error]\u001b[0m Wrong config file type ({config_file}).")
            exit()

        lines = lines[1:]

        self.plane_file = lines[0].split(": ")[1:][0].strip()
        self.wing_aerofoil = lines[1].split(": ")[1:][0]
        self.elevator_aerofoil = lines[2].split(": ")[1:][0]
        self.fin_aerofoil = lines[3].split(": ")[1:][0]
        self.angle_min = float(lines[4].split()[1])
        self.angle_max = float(lines[5].split()[1])
        self.increment = int(lines[6].split()[1])
        self.span_loc = float(lines[7].split()[1])
        self.threads = float(lines[8].split()[1])
        self.show_geom_plt = str_to_bool(lines[9].split(": ")[1][0])

        return None

    def generate_planes(self):
        """
        Generates tail configurations, saves as AVL readable plane files.

        Returns:
            planes {list[Plane]} -- List of plane objects with modified geometry.
        """
        #   Generate reference plane goemetry. Strips wing section to be modified and removes fin.
        self.ref_plane = Plane(name="REF")
        self.ref_plane.read(self.plane_file)
        self.ref_plane.strip_section("Main Wing")
        self.ref_plane.strip_surface("Fin")

        planes = []

        mac = self.ref_plane.mac
        span = self.ref_plane.b_w
        # Half span (AVL wings are defined from centreline to outboard.)
        hspan = span/2

        count = 0
        theta_range = np.linspace(  # Dihedral angle range.
            self.angle_min,
            self.angle_max,
            int(1+(self.angle_max-self.angle_min)/self.increment)
        )
        for theta in theta_range:
            name = str(count)

            plane = Plane(name=name)
            plane.dihedral_angle = theta
            plane.dihedral_split = self.span_loc

            # Location to split wing for dihedral start.
            split_loc = hspan*self.span_loc/100
            plane.dihedral_splitY = split_loc
            plane.span = span

            # Copy required because reasons.
            mod_geom = copy.copy(self.ref_plane.file_str)

            # Calculates tip Z due to dihedral angle
            Zle = round((hspan-split_loc)*np.sin(np.radians(theta)), 3)
            plane.tipZ = Zle
            # Calcualtes tip Y due to dihedral angle
            Yle = round((hspan-split_loc) *
                        np.cos(np.radians(theta))+split_loc, 3)
            plane.tipY = Yle

            #   Generate root, split and tip sections in AVL format.
            root = Section(self.ref_plane.Xle, 0, 0, mac, int(
                split_loc*0.02), -2, self.elevator_aerofoil)
            split = Section(self.ref_plane.Xle, split_loc, 0, mac, int(
                (np.sqrt(Yle**2+Zle**2)-split_loc)*0.02), -1, self.elevator_aerofoil)
            # Creates tip section based off tip geometry
            tip = Section(self.ref_plane.Xle, Yle, Zle,
                          mac, 0, 0, self.elevator_aerofoil)

            mod_str = str(root)  # Gets section string
            mod_str += str(split)
            mod_str += str(tip)

            for index, line in enumerate(mod_geom):
                if line == "MARKER\n":  # Finds marker
                    mod_geom.pop(index)  # Removes marker
                    # Inserts modified sections
                    mod_geom.insert(index, mod_str)

            #   Writes plane file.
            file_name = name = f"{plane.name}-{theta}deg-{self.span_loc}%"
            plane.geom_file = f"{self.path}/generated planes/{file_name}.avl"
            with open(plane.geom_file, 'w') as file:
                file.write("".join(mod_geom))
            count += 1

            planes.append(plane)

        self.planes = planes

        return planes

    def run(self):
        """
        Runs aero analysis.
        """
        aero = Aero(self.aero_config_file)  # initialises aero analysis, reads config file.
        if aero.polars == False:
            raise ValueError("Polars must be enabled for dihedral analysis.")

        #   Can't do multithreaded analysis without some thinking and extra code :(
        for plane in tqdm(self.planes, desc="Aero analysis"):
            aero.run(plane)

        return None

    def plot(self):
        """
        Main plot function. Handles polar and eigenmode plots in subplots.
        """
        fig, (ax1, ax2, ax3) = plt.subplots(
            ncols=3, figsize=(12, 3), sharex=True)

        polar_plt = self.plot_polars(ax1, ax2, ax3)
        # mode_plt=self.plot_modes(ax4)

        plt.tight_layout()

        if self.plot_dihedral_angle == True:
            geom_plt = self.plot_dihedral_angle()

        plt.show()

        return None

    def plot_polars(self, ax1, ax2, ax3):
        """
        Draws polar plots.

        Arguments:
            ax1, ax2, ax3 {matplotlib.Axes} -- Subplot axes.

        Returns:
            plt {matplotlib.pyplot}
        """
        dihedral_angles = [plane.dihedral_angle for plane in self.planes]

        #   Aero polar plot
        Cl_0 = self.planes[0].polars['Cl'].iloc[-1]
        Cd_0 = self.planes[0].polars['Cd'].iloc[-1]
        Cl_delta = [100*(plane.polars['Cl'].iloc[-1]-Cl_0) /
                    Cl_0 for plane in self.planes]
        Cd_delta = [100*(plane.polars['Cd'].iloc[-1]-Cd_0) /
                    Cd_0 for plane in self.planes]

        ax1.plot(dihedral_angles, Cl_delta, label="Lift ($C_{L}$)")
        ax1.plot(dihedral_angles, Cd_delta, label="Lift ($C_{D}$)")

        ax1.set_ylabel(
            f"\u0394 (%) @ {self.planes[0].polars['Alpha (deg)'].iloc[-1]}\u00B0 AoA")
        ax1.legend(loc='upper left')
        ax1.set_title("Aero Coeffients")

        #   Stability derivative plot.
        Clb = [plane.polars['Clb'].iloc[0] for plane in self.planes]
        Clp = [plane.polars['Clp'].iloc[0] for plane in self.planes]

        ax2.plot(dihedral_angles, Clb, label="Dihedral ($Cl_{b}$)")
        ax2.plot(dihedral_angles, Clp, label="Roll Rate ($Cl_{p}$)")

        ax2.legend()
        ax2.set_title("Stability Derivatives")
        ax2.set_ylabel("Dervative [NA]")
        ax2.set_xlabel(
            f"Dihedral Angle (\u00B0) - Split Location={self.planes[0].dihedral_split}% of Span")

        #   Spiral stability plot
        spiral = [plane.polars['spiral'].iloc[1] for plane in self.planes]

        ax3.plot(dihedral_angles, spiral)
        ax3.set_title("Spiral Stability (>1 = stable)")
        # ax3.set_xlabel(f"Dihedral Angle (\u00B0) - Split Location={self.planes[0].dihedral_split}% of Span")
        ax3.set_ylabel("Clb.Cnr / Clr.Cnb [NA]")

        return plt

    def plot_modes(self, ax4):
        """
        Draws eigenmode plot.

        Arguments:
            ax4 {matplotlib.Axes} -- Subplot axes.

        Returns:
            plt {matplotlib.pyplot}
        """
        dihedral_angles = [plane.dihedral_angle for plane in self.planes]

        roll_0 = self.planes[0].modes["roll"][0][0]
        dutch_0 = self.planes[0].modes["dutch"][0][0]
        roll_delta = [100*(plane.modes["roll"][0][0]-roll_0) /
                      roll_0 for plane in self.planes]
        dutch_delta = [100*(plane.modes["dutch"][0][0] -
                            dutch_0)/dutch_0 for plane in self.planes]

        ax4.set_xlabel(
            f"Dihedral Angle (\u00B0) - Split Location={self.planes[0].dihedral_split}% of Span")
        ax4.set_ylabel("\u0394 Damping (%)")
        ax4.set_title("Eigenmode Damping")

        ax4.plot(dihedral_angles, roll_delta, label="Roll")
        ax4.plot(dihedral_angles, dutch_delta, label="Dutch Roll")
        ax4.legend()

        return

    def plot_dihedral_angle(self):
        """
        Draws spanwise dihedral geometry plot.

        Returns:
            plt {matplotlib.pyplot}
        """
        plt.figure()
        plt.title("Spanwise Geometry Plot")

        plt.xlabel("Y (mm)")
        plt.ylabel("Z (mm)")
        plt.xlim(0, max([plane.tipY for plane in self.planes]))
        plt.ylim(0, max([plane.tipY for plane in self.planes]))

        for plane in self.planes:
            plt.plot([0, plane.dihedral_splitY, plane.tipY],
                     [0, 0, plane.tipZ])

        return plt


if __name__ == "__main__":
    dihedral = Dihedral('dihedral.config', 'aero.config')
    dihedral.generate_planes()
    dihedral.run()
    dihedral.plot()



import argparse
import os

from .aero import Aero
from .geometry import Plane
from .dihedral import Dihedral
from .tail import AutoTail

parser=argparse.ArgumentParser(description="AVL Automation.")

parser.add_argument('run_type',choices=['aero','tail','dihedral'],help='Type of analysis to run.')
parser.add_argument('-p','--plane',action='store',help="Plane .avl file for aero analysis.")
parser.add_argument('-c','--config',nargs='+',action='store',help="Config file for analysis.")

args=parser.parse_args()

if args.run_type=='aero':
    if args.plane is None:
        parser.error("Aero requires --plane.")

    if args.config is None:
        parser.error("Aero requires --config.")

    if len(args.config)>1:
        parser.error("Aero requires only 1 config file.")

    if os.path.exists(args.config[0])==False:
        print(f"\u001b[31m[Error]\u001b[0m {args.config[0]} not found.")
        exit()

    plane=Plane(geom_file=args.plane)

    aero=Aero(args.config[0])
    aero.run(plane)

    if aero.polars==True:
        print('\nPolars:\n',plane.polars)
    if aero.modes==True:
        print('\nEigenmodes:\n',plane.modes,'\n')

if args.run_type=='tail':
    if args.config is None:
        parser.error("Tail requires --config.")
        
    if len(args.config)>1:
        parser.error("Tail requires only 1 config file.")

    if os.path.exists(args.config[0])==False:
        print(f"\u001b[31m[Error]\u001b[0m {args.config[0]} not found.")
        exit()
    
    tail=AutoTail(args.config[0])
    tail.generate_planes()
    tail.run()
    tail.results()

if args.run_type=='dihedral':
    if args.config is None:
        parser.error("Dihedral requires --config.")
    
    if len(args.config)!=2:
        parser.error("Dihedral requires 2 config files: dihedral, aero.")

    for config in args.config:
        if os.path.exists(config)==False:
            print(f"\u001b[31m[Error]\u001b[0m {config} not found.")
            exit()

    
    dihedral=Dihedral(args.config[0],args.config[1])
    dihedral.generate_planes()
    dihedral.run()
    dihedral.plot()


class KeyErrorMessage(str):
    def __repr__(self): return str(self)


class Plane():
    def __init__(self,name:str=None,geom_file=None):

        self.name=name
        self.geom_file=geom_file
        self.results_file=None
        self.Xcg=None
        self.np=None
        self.sm=None
        self.sm_ideal=None
        self.Xt=None
        self.Lt=None
        self.Sw=None
        self.Ct_v=None
        self.St_h=None  #   Equivilent horizontal tail area
        self.St_v=None  #   Equivilent vertical tail area
        self.mac=None
        self.ARh=None
        self.ARv=None
        self.dihedral_angle=None
        self.dihedral_split=None
        self.dihedral_splitY=None
        self.tipY=None
        self.tipZ=None
        self.Xle=None
        self.cases=None
        self.polars=None
        self.modes=None
        self.tail_config=None
        self.b_w=None
        self.b_th=None
        self.b_tv=None
        self.c_t=None
        self.theta=None
        self.Xw_root=None
        self.Cw_root=None

        if self.name==None:
            self.name="plane"

        if geom_file!=None:
            self.read(geom_file)

        return None

    def read(self,file:str):
        """
        Loads AVL plane file.

        Parameters:
        -----------
        file: str; AVL plane geometry file.

        Returns:
        -------
        file_str: list; A list of lines in the plane file.
        """
        file_str=list()
        with open(file,'r') as f:
            lines=f.readlines()

        wing=False

        for i,line in enumerate(lines):
            if line=="\n" or line[0]=="#":  #   Ignores blank lines & comments
                continue
            else:
                file_str.append(line)

            if line.strip()=="Main Wing":
                wing=True
            if line.split()[0]=="SECTION" and wing==True:
                Xle=lines[i+1].split()[0]
                wing=False

        self.Xle=Xle

        ref_dims=[n for n in file_str][3]
        self.Sw=float(ref_dims.split()[0])
        self.mac=float(ref_dims.split()[1])
        self.b_w=float(ref_dims.split()[2])
        self.ARw=self.b_w/self.mac

        self.file_str=file_str

        ### Get wing LE x location
        surface=False
        section=False
        found=False
        for i,line in enumerate(self.file_str):
            if line.strip()=="Main Wing":
                found=True
                surface=True
            if line.split()[0]=="SECTION" and surface==True:
                section=True
                self.Xw_root=float(self.file_str[i+1].strip().split()[0])
                self.Cw_root=float(self.file_str[i+1].strip().split()[3])

                break

        if found!=True:
            msg=KeyErrorMessage("Wing surface not found. Wing should be defined by:\n\tSURFACE\n\tMain Wing\n\t...")
            raise KeyError(msg)

        return None

    def strip_section(self,section_name):   
        """
        Removes section by name.

        Parameters:
        ----------
        section: str; Name of section in plane file to remove.

        Returns:
        --------
        None - section is stripped inplace.
        """
        stripped_str=[]

        surface=False
        section=False
        found=False
        for line in self.file_str:
            if line.strip()==section_name:
                found=True
                surface=True
            if line.split()[0]=="SECTION" and surface==True:
                section=True
            if section!=True:   #   Adds everything that isn't section 
                stripped_str.append(line)
            if line.split()[0]=="SURFACE" and surface==True:    #   Finds next surface
                stripped_str.append("MARKER\n")    #   Adds marker for adding new sections
                section=False
                surface=False
                stripped_str.append(line)  #   Adds the rest of the file
        
        self.file_str=stripped_str

        if found==False:
            stripped_str.append(f"""\nSURFACE
{section_name}
13 1.0
INDEX
937
YDUPLICATE
0.0
SCALE
1.0 1.0 1.0
TRANSLATE
0.0 0.0 0.0\n""")
            stripped_str.append("MARKER\n")
            raise KeyError(f"Section '{section_name}' not found.")

        return None

    def strip_surface(self,surface_name):   
        """
        Removes surface by name.

        Parameters:
        ----------
        section: str; Name of surface in plane file to remove.

        Returns:
        --------
        None - surface is stripped inplace.
        """
        stripped_str=[]

        surface=False
        found=False
        for i,line in enumerate(self.file_str):
            if line.split()[0]=="SURFACE":
                if self.file_str[i+1].split()[0]==surface_name:
                    found=True
                    surface=True
                if self.file_str[i+1].split()[0]!=surface_name:
                    surface=False

            if surface==False:
                stripped_str.append(line)

        self.file_str=stripped_str

        if found==False:
            raise KeyError(f"Surface '{surface_name}' not found.")

        return None
    
    def calc_SM(self):
        """
        Reads stability analysis results file and calculates SM based
        on MAC, neutral point, and Xcg.

        Returns:
        sm: float; Static margin.
        """
        with open(self.results_file,'r') as text:
            lines=text.readlines()[50]

        self.np=float(lines.split()[-1])
        self.sm=(self.np-self.Xcg)/self.mac

        return self.sm

    def calc_Xcg_ideal(self):
        """
        Reads stability analysis results file and calculates ideal Xcg
        based on MAC, neutral point, and ideal SM.

        Returns:
        Xcg: float; Ideal CG location in X.
        """
        with open(self.results_file,'r') as text:
            lines=text.readlines()[50]

        self.np=float(lines.split()[-1])
        self.Xcg=self.np-(self.mac*self.sm_ideal)

        return self.Xcg

class Surface():
    #Creates surface (eg wing type)
    def __init__(self,name,nchord,cspace,component,aerofoil,y_duplicate=None,angle=None):
        self.name=name
        self.nchord=nchord
        self.cspace=cspace
        self.component=component
        self.y_duplicate=y_duplicate
        self.angle=angle
        self.aerofoil=aerofoil

    def __str__(self):
        surf_str    =   f"{self.name}\n#Nchord Cspace\n"
        surf_str    +=  f"{self.nchord} {self.cspace}\n"

        if self.y_duplicate is not None:
            surf_str+=  f"COMPONENT\n{self.component}\n"

        surf_str    +=  f"YDUPLICATE\n{self.y_duplicate}\n"
        surf_str    +=  f"SCALE\n1 1 1\n"
        surf_str    +=  f"TRANSLATE\n0 0 0\n"

        if self.angle is not None:
            surf_str+=  f"ANGLE\n{self.angle}\n"

        return surf_str

class Section():
    def __init__(self,Xle,Yle,Zle,chord,nspan,sspace,aerofoil):
        self.Xle=Xle
        self.Yle=Yle
        self.Zle=Zle
        self.chord=chord
        self.nspan=nspan
        self.sspace=sspace
        self.aerofoil=aerofoil

    def __str__(self):
        section_str="SECTION\n#Xle Yle Zle Chord Ainc Nspan Sspace\n"
        section_str+=f"{self.Xle} {self.Yle} {self.Zle} {self.chord} {0} {self.nspan} {self.sspace}\n"
        section_str+=f"AFIL 0.0.1.0\n{self.aerofoil}\n"

        return section_str

if __name__=="__main__":
    plane=Plane('aria3','Aria3.avl')



import os
import shutil
import numpy as np
from matplotlib import cm
from matplotlib import pyplot as plt
from concurrent.futures import ThreadPoolExecutor
from tqdm import tqdm
import copy
import pandas as pd
from scipy import optimize

from .geometry import Plane, Section
from .aero import Case, avl_cmd


class CurveFit():
    def __init__(self, planes: list[Plane], sm_ideal: float):
        self.planes = planes
        self.sm_ideal = sm_ideal

        self.Lts = np.array([plane.Lt for plane in self.planes])
        self.Xts = np.array([plane.Xt for plane in self.planes])
        self.SMs = np.array([plane.sm for plane in self.planes])
        self.St_hs = np.array([plane.St_h for plane in self.planes])

        if self.SMs.min() > sm_ideal or self.SMs.max() < sm_ideal:
            print(
                "\u001b[33m[Warning]\u001b[0m No stable configurations found. Consider changing limits.")

            self.unstable = True
        else:
            self.unstable = False

    def func(self, data: np.ndarray, a: float, b: float, c: float, d: float) -> np.ndarray:
        """Equation for 3D surface (applicable to Lt, Sh, SM datapoints)

        Args:
            data (np.ndarray): x,y data.
            a (float): Surface control parameter.
            b (float): Surface control parameter.
            c (float): Surface control parameter.
            d (float): Surface control parameter.

        Returns:
            np.ndarray: z values.
        """
        x = data[0]
        y = data[1]
        z = a*(x**b)*(y**c)+d
        return z

    def func_inv_const_z(self, x: np.ndarray, z: float, a: float, b: float, c: float, d: float) -> np.ndarray:
        """Solve 'func' for y given x and z.

        Args:
            x (np.ndarray): x data.
            z (float): z value at which to solve func.
            a (float): Surface control parameter.
            b (float): Surface control parameter.
            c (float): Surface control parameter.
            d (float): Surface control parameter.

        Returns:
            np.ndarray: y values.
        """
        y = np.exp((1/c)*np.log((z-d)/(a*x**b)))
        return y
    
    def Lt_to_Xt(self, x):
        return np.interp(x, self.Lts, self.Xts)

    def Xt_to_Lt(self, x):
        return np.interp(x, self.Xts, self.Lts)

    def curve_fit(self, x: np.ndarray, y: np.ndarray, z: np.ndarray) -> np.ndarray:
        """Scipy curve fit for func.

        Args:
            x (np.ndarray): Intiial datapoints.
            y (np.ndarray): Intiial datapoints.
            z (np.ndarray): Intiial datapoints.

        Returns:
            np.ndarray: Surface control parameters.
        """

        parameters, covariance = optimize.curve_fit(self.func, [x, y], z)

        return parameters

    def curve_fit_slice(self) -> list[plt.figure, plt.axes]:
        """
        Slices surface fit to AVL datapoints.

        Returns:
            Lt: {np.ndarray} -- Tail moment arm
            St_h: {np.ndarray} -- Horizontal tail area
            St_v: {np.array} -- Vertical tail area

        """

        if self.unstable == True:
            print("\u001b[31m[Error]\u001b[0m SM ideal is out of range of analysis datapoints. Stable configurations are required to slice at SM ideal.")
            exit()

        Lts = self.Lts
        SMs = self.SMs
        St_hs = self.St_hs

        parameters = self.curve_fit(St_hs, Lts, SMs)

        St_h = np.linspace(St_hs.min(), St_hs.max(), 20)
        Lt = self.func_inv_const_z(St_h, self.sm_ideal, *parameters)
        St_v = self.planes[0].Ct_v*self.planes[0].Sw*self.planes[0].b_w/Lt

        return Lt, St_h, St_v

    def curve_fit_surface(self) -> list[np.ndarray]:
        """Fits surface to Lt, St_h, SM datapoints.

        Returns:
            list[np.ndarray]: x,y,z data (Lt, St_h, SM)
        """

        St_hs = [plane.St_h for plane in self.planes]
        Lts = self.Lts
        Sms = [plane.sm for plane in self.planes]

        parameters = self.curve_fit(self.Lts, self.St_hs, self.SMs)

        St_h_range = np.linspace(min(self.St_hs), max(self.St_hs), 20)
        Lt_range = np.linspace(min(self.Lts), max(self.Lts), 20)

        x2, y2 = np.meshgrid(Lt_range, St_h_range)
        z2 = self.func(np.array((x2, y2)), *parameters)

        return x2, y2, z2

    def plot_slice(self, Lt: np.ndarray, St_h: np.ndarray, St_v: np.ndarray):
        """
        Plots slices.

        Returns:
            fig: {plt.figure}
            ax1: {plt.axes}
            ax2: {plt.axes}
        """
        fig, ax1 = plt.subplots(figsize=(7, 7))

        ax1.plot(Lt, St_h, color='r', linestyle='-',
                 label=fr"Horizontal Tail"
                 )
        ax1.plot(Lt, St_v, color='b', linestyle='-',
                 label=fr"Vertical Tail"
                 )

        ax1.set_ylabel(r"$St$ ($Lunit^2$)")
        ax1.set_xlabel(r"$Lt$ ($Lunit$)")
        if max((max(St_h), max(St_v))) > 1000:
            ax1.ticklabel_format(axis='y', style='sci', scilimits=(0, 0))
        ax1.legend()

        ax2 = ax1.secondary_xaxis(
            'top', functions=(self.Lt_to_Xt, self.Xt_to_Lt))
        ax2.set_xlabel(r"Xt ($Lunits$)")

        ax1.set_title(fr"Tail Configurations with $SM={self.sm_ideal}$")
        # fig.tight_layout()

        return fig, ax1, ax2

    def plot_surface(self, x, y, z):
        fig = plt.figure()
        ax = fig.add_subplot(projection='3d')

        ax.plot_surface(x, y, z, cmap=cm.viridis)
        ax.scatter(self.Lts, self.St_hs, self.SMs, color='k', depthshade=False)

        ax.set_xlabel("${St_h}$ (${Lunit^2}$)")
        ax.set_ylabel("${Lt}$ (${Lunit}$)")
        ax.set_zlabel("SM")

        fig.tight_layout()

        return plt

    def plot_surface_contour(self, x, y, z):

        fig, ax = plt.subplots()

        cs = ax.contour(x, y, z, 10, colors='k')
        ax.clabel(cs, cs.levels, inline=True, colors='k')

        ax.set_xlabel("${St_h}$ (${Lunit^2}$)")
        ax.set_ylabel("${Lt}$ (${Lunit}$)")

        fig.tight_layout()

        return plt


class AutoTail():
    def __init__(self, config_file: str):
        
        self.path = os.path.split(config_file)[0]
        
        try:
            if os.path.isdir(self.path+"/results") == True:
                shutil.rmtree(self.path+"/results")
            if os.path.isdir(self.path+"/generated planes") == True:
                shutil.rmtree(self.path+"/generated planes")
            if os.path.isdir(self.path+"/cases") == True:
                shutil.rmtree(self.path+"/cases")
        except PermissionError:
            raise PermissionError("Close all results, geometry, case files")

        os.mkdir(self.path+"/generated planes")
        os.mkdir(self.path+"/results")
        os.mkdir(self.path+"/cases")
        
        if os.path.exists(f"{self.path}/avl.exe")==False:
            print("\u001b[31m[Error]\u001b[0m avl.exe not found.")
            exit()

        self.read_config(config_file)

        return None

    def read_config(self, file: str):
        """Reads tail.config data. 

        Args:
            file (str): tail.config file path.
        """

        with open(file, 'r') as f:
            lines = [line for line in f.readlines()]
        lines = [line for line in lines if line[0] != "#" and line != "\n"]

        if lines[0].strip() != "TAIL CONFIG":
            print(
                f"\u001b[31m[Error]\u001b[0m Wrong config file type ({file}).")
            exit()
        lines = lines[1:]

        self.plane_file = "".join(
            (self.path, '/', lines[0].split(": ")[1:][0].strip()))
        self.wing_aerofoil = "".join(
            (self.path, '/', lines[1].split(": ")[1:][0]))
        self.elevator_aerofoil = "".join(
            (self.path, '/', lines[2].split(": ")[1:][0]))
        self.fin_aerofoil = "".join(
            (self.path, '/', lines[3].split(": ")[1:][0]))

        self.Xcg = lines[4].split()[1]
        self.Ycg = lines[5].split()[1]
        self.Zcg = lines[6].split()[1]
        self.mass = float(lines[7].split()[1])

        self.Xt_upper = float(lines[8].split()[1])
        self.Xt_lower = float(lines[9].split()[1])
        self.St_h_upper = float(lines[10].split()[1])
        self.St_h_lower = float(lines[11].split()[1])
        self.Ct_v = float(lines[12].split()[1])
        self.steps = 7

        self.sm_ideal = float(lines[13].split()[1])
        self.tolerance = float(lines[14].split()[1])
        self.config = float(lines[15].split()[1])
        self.b_th = lines[16].split()[1]

        self.threads = int(lines[17].split()[1])

        if self.b_th != "NA":
            self.b_th = float(self.b_th)
        if self.b_th == 0:
            print(
                "\u001b[31m[Error]\u001b[0m Input non-zero horizontal span constraint. (NA to ignore)")
            exit()
        if self.b_th != "NA" and self.config == 0:
            print(
                "\u001b[33m[Warning]\u001b[0m Span constraint will be ignored for conventional tails.")

        if self.Xt_lower == 0 or self.St_h_lower == 0:
            print("\u001b[31m[Error]\u001b[0m Input non-zero lower bound.")
            exit()

        if self.Xcg == "NA" and self.Ycg == "NA" and self.Zcg == "NA":
            self.calc_cg = True
        else:
            self.calc_cg = False

            self.Xcg = float(self.Xcg)
            self.Ycg = float(self.Ycg)
            self.Zcg = float(self.Zcg)

        if self.config != 0 and self.config != 1:
            print(
                "\u001b[31m[Error]\u001b[0m Invalid tail configuration selected.")
            exit()

        self.St_h_range = np.linspace(
            self.St_h_lower, self.St_h_upper, self.steps)
        self.Xt_range = np.linspace(self.Xt_lower, self.Xt_upper, self.steps)

    def generate_planes(self):
        """Generates planes according to user tail limits. Generates and writes AVL geometry file.

        Returns:
            List[Plane]: List of Plane generated plane objects.
        """
        self.ref_plane = Plane(name="REF")
        self.ref_plane.read(self.plane_file)
        try:
            self.ref_plane.strip_section("Elevator")
        except KeyError:
            print(
                "\u001b[33m[Warning]\u001b[0m No section 'Elevator' found. Check if geometry of generated planes looks correct.")
        try:
            self.ref_plane.strip_surface("Fin")
        except KeyError:
            print(
                "\u001b[33m[Warning]\u001b[0m No surface 'Fin' found. Check if geometry of generated planes looks correct.")

        planes = []

        self.Sw = self.ref_plane.Sw
        self.mac = self.ref_plane.mac
        self.b_w = self.ref_plane.b_w
        ARw = self.ref_plane.ARw
        ARh = ARw*2/3
        Xw_root = self.ref_plane.Xw_root
        Cw_root = self.ref_plane.Cw_root

        count = 0
        for St_h in self.St_h_range:
            for Xt in self.Xt_range:
                St_h = float(St_h)
                Xt = Xt

                name = str(count)  # Creates plane name
                plane = Plane(name=name)  # Initializes new plane

                plane.Xt = Xt
                plane.Sw = self.Sw
                plane.Xw_root = Xw_root
                plane.Cw_root = Cw_root
                plane.St_h = St_h
                plane.ARh = ARh
                plane.mac = self.mac
                plane.b_w = self.b_w
                plane.sm_ideal = self.sm_ideal
                plane.tail_config = self.config
                plane.Ct_v = self.Ct_v

                if self.calc_cg == False:
                    plane.Xcg = self.Xcg

                mod_geom = copy.copy(self.ref_plane.file_str)

                if self.b_th != "NA" and self.config == 1:  # if span constraint used:
                    chord = St_h/self.b_th  # Calculate chord based off span & area, not area & AR
                    span = self.b_th
                else:
                    # Calculates h chord based on area & AR
                    chord = np.sqrt(St_h/plane.ARh)
                    # Calculates HTP span (Lunit)
                    span = np.sqrt(St_h*plane.ARh)

                plane.b_th = span
                plane.c_t = chord

                plane.Lt = (plane.Xt+plane.c_t*0.25) - \
                    (plane.Xw_root+0.25*plane.Cw_root)
                if plane.Lt <= 0:
                    print(
                        "\u001b[31m[Error]\u001b[0m Tail moment arm <=0. Increase Xt lower bound.")
                    exit()

                plane.St_v = plane.Ct_v*plane.Sw*plane.b_w/plane.Lt  # Vertical tail sizing

                # Calculates tip height (inverted v tail) (Lunit)
                Zle = (plane.St_v)/(2*chord)
                plane.theta = np.rad2deg(np.arctan(Zle/(span/2)))

                if self.config == 0:
                    # Defines root section (object)
                    root = Section(Xt, 0, 0, chord, 10, -1,
                                   self.elevator_aerofoil)
                elif self.config == 1:
                    root = Section(Xt, 0, Zle, chord, 10, -
                                   1, self.elevator_aerofoil)

                # Defines tip section (object)
                tip = Section(Xt, span/2, 0, chord, 10, -
                              2, self.elevator_aerofoil)
                # Combines 2 sections to insert into reference plane
                mod_str = str(root)+str(tip)

                for index, line in enumerate(mod_geom):
                    if line == "MARKER\n":
                        mod_geom.pop(index)  # Removes marker
                        # Inserts modified sections
                        mod_geom.insert(index, mod_str)

                file_name = f"{plane.name}-{str(round(St_h,2))}Sh-{str(round(plane.Lt,2))}Lt"
                plane.geom_file = f"{self.path}/generated planes/{file_name}.avl"

                with open(plane.geom_file, 'w') as file:
                    file.write("".join(mod_geom))
                count += 1

                planes.append(plane)

        print("[Info] Planes generated.")
        self.planes = planes

        return planes

    def run(self):
        """Runs AVL stability analysis. Multithreaded due to high io throughput.
        """
        self.case = Case(self.path,self.Xcg, self.Ycg, self.Zcg, self.mass)
        self.case.write_stab_case()

        tasks = [(self.case, plane) for plane in self.planes]
        # Starts analysis on multiple threads
        with ThreadPoolExecutor(max_workers=self.threads) as pool:
            list(tqdm(pool.map(self.stab_analysis, tasks),
                 total=len(tasks), desc="Stability analysis"))

        tasks = [plane for plane in self.planes]
        # Starts post processing on multiple threads
        with ThreadPoolExecutor(max_workers=self.threads) as pool:
            pool.map(self.calc_SM, tasks)

    def stab_analysis(self, tasks):
        """Creates AVL input string and executes AVL analysis.

        Args:
            tasks (List): Case and Plane to run [Case, Plane].

        Returns:
            None
        """
        case, plane = tasks

        cmd_str = "load {0}\n".format(plane.geom_file)  # Load plane
        cmd_str += "case {0}\n".format(case)  # Load case
        cmd_str += "oper\n x\n"  # Run analysis
        cmd_str += "st\n"  # View stability derivatives

        plane.results_file = f"{self.path}/results/"+plane.name+".txt"
        cmd_str += plane.results_file+"\n"  # Saves results
        
        avl_cmd(cmd_str, self.path)

    def calc_SM(self, tasks):
        """Calculates static margin for each plane.

        Args:
            tasks (Plane)
        """
        plane = tasks
        if self.calc_cg == False:
            plane.calc_SM()
        else:
            plane.calc_Xcg_ideal()

    def results(self, display=True):
        """Collates results.

        Args:
            display (bool, optional): _description_. Defaults to True.

        Returns:
            pd.DataFrame: solutions dataframe (redundant).
            CurveFit: CurveFit object for stable tail configs.
        """

        if self.calc_cg == False:

            columns = ["Plane ID", "Static Margin", "Xnp (Lunit)", "Xt (Lunit)", "Lt (Lunit)",
                       "Span (Lunit)", "Chord (Lunit)", "Angle (deg)", "Sh (Lunit^2)", "Sv (Lunit^2)", "ARh"]
            solutions = []
            for plane in self.planes:
                if np.isclose(plane.sm, plane.sm_ideal, atol=self.tolerance) == True:
                    solutions.append([
                        plane.name.split("-")[0],
                        plane.sm,
                        plane.np,
                        plane.Xt,
                        plane.Lt,
                        plane.b_th,
                        plane.c_t,
                        plane.theta,
                        plane.St_h,
                        plane.St_v,
                        plane.ARh,
                    ])

            solutions_df = pd.DataFrame(solutions, columns=columns)
            solutions_df = solutions_df.round(2)

            if self.config == 0:
                solutions_df = solutions_df[[
                    "Plane ID", "Static Margin", "Xnp (Lunit)", "Xt (Lunit)", "Lt (Lunit)", "Sh (Lunit^2)", "Sv (Lunit^2)", "ARh"]]

            curve_fit = CurveFit(self.planes, self.sm_ideal)
            if curve_fit.unstable==False:
                print(
                    "Consider refining limits around possible configurations.\n")

            print(f"np: {self.Xcg-(self.planes[0].mac*self.sm_ideal)} Lunit")

            if display == True:
                ##### Generated planes SM results (3D plot) #####
                x2, y2, z2 = curve_fit.curve_fit_surface()
                curve_fit.plot_surface(x2, y2, z2)

                if curve_fit.unstable==False:
                    Lt, St_h, St_v = curve_fit.curve_fit_slice()
                    curve_fit.plot_slice(Lt, St_h, St_v)


                plt.show()

            return solutions_df, curve_fit

        elif self.calc_cg == True:

            columns = ["Plane ID", "Xcg (Lunit)",
                       "np (Lunit)", "Static Margin"]
            solutions = []
            for plane in self.planes:
                solutions.append(
                    [plane.name.split("-")[0], plane.Xcg, plane.np, self.sm_ideal])

            solutions_df = pd.DataFrame(solutions, columns=columns)
            solutions_df = solutions_df.round(2)

            if display == True:
                fig = plt.figure()
                ax = fig.add_subplot(projection='3d')

                x = [plane.St_h for plane in self.planes]
                y = [plane.Lt for plane in self.planes]
                z = [plane.np for plane in self.planes]

                ax.scatter(x, y, z)

                ax.set_xlabel("St_h (Lunit^2)")
                ax.set_ylabel("Xt")
                ax.set_zlabel(f"Xcg for SM={self.planes[0].sm_ideal}")

                print("\n", solutions_df)

                plt.show()

            return solutions_df


if __name__ == "__main__":
    tail = AutoTail("projects/tail_MDDP_v0.config")
    tail.generate_planes()
    tail.run()
    tail.results()



import subprocess as sp
import numpy as np
from concurrent.futures import ThreadPoolExecutor
import pandas as pd
import os
import shutil

from .geometry import Plane

def avl_cmd(cmd_str:str,path:str)->None:
    """
    Opens AVL in a subprocess and submits command string.

    Arguments:
        cmd_str {string} -- Command string to be submitted to AVL. Essentially key presses.
    """

    avl_subprocess=sp.Popen(
        [f"{path}/avl.exe"],
        stdin=sp.PIPE,
        stdout=sp.PIPE,
        stderr=sp.PIPE
    )

    avl_subprocess.communicate(input=cmd_str.encode())

    return None

class Case():
    def __init__(self,path,Xcg,Ycg,Zcg,mass,Ixx=None,Iyy=None,Izz=None,velocity=None,density=None,alpha=None,modes=False,polars=False,id=False):
        """
        Most of these class inits were written when I didn't fully understand what they were for lol
        """
        
        self.path=path
        self.Xcg=Xcg
        self.Ycg=Ycg
        self.Zcg=Zcg
        self.Ixx=Ixx
        self.Iyy=Iyy
        self.Izz=Izz
        self.mass=mass
        self.velocity=velocity
        self.density=density
        self.alpha=alpha
        self.case_file=None
        self.Cl=None
        self.Cd=None
        self.modes_results_file=None
        self.polars_results_file=None
        self.modes=modes
        self.polars=polars
        self.id=id
        self.Clb=None
        self.Clp=None
        self.spiral=None

    def write_aero_case(self):
        """
        Creates aero polar case string in AVL format and writes to file.
        """
        case_str =  "\n---------------------------------------------\n"
        case_str += f"Run case  1:\n\n"
        case_str += f"alpha -> alpha = {self.alpha}\n"
        case_str += f"X_cg={self.Xcg} Lunit\n"
        case_str += f"Y_cg={self.Ycg} Lunit\n"
        case_str += f"Z_cg={self.Zcg} Lunit\n"
        case_str += f"mass={self.mass} kg\n"
        case_str += f"Ixx={self.Ixx} kg-m^2\n"
        case_str += f"Iyy={self.Iyy} kg-m^2\n"
        case_str += f"Izz={self.Izz} kg-m^2\n"
        case_str += f"velocity={self.velocity} m/s\n"
        case_str += f"density={self.density} kg-m^3\n"
        case_str += "grav.acc.=0.98 m/s^2\n"

        path=f"{self.path}/cases/{str(self.alpha)}deg.case"

        with open(path,'w') as f:
            f.write(case_str)

        self.case_file=path

        return None

    def write_stab_case(self)->None:
        """
        Creates stability case string & writes to file
        """
        case_str =  "\n---------------------------------------------\n"
        case_str += "Run case  1:\n\n"
        case_str += f"X_cg={self.Xcg} Lunit\n"
        case_str += f"Y_cg={self.Ycg} Lunit\n"
        case_str += f"Z_cg={self.Zcg} Lunit\n"
        case_str += f"mass={self.mass} kg\n"

        path=f"{self.path}/cases/tail.case"
        with open(path,'w') as file:    #   Saves case file
            file.write(case_str)

        self.case_file=path

        return None 

class Aero():
    def __init__(self,config_file:str):
        #   Cleans temp folders.
        self.path = os.path.split(config_file)[0]

        if os.path.isdir(self.path+"/cases")==True:
            shutil.rmtree(self.path+"/cases")
        os.mkdir(self.path+"/cases")
        if os.path.isdir(self.path+"/results")==True:
            shutil.rmtree(self.path+"/results")
        os.mkdir(self.path+"/results")

        self.read_config(config_file)

        alpha_range=np.linspace(    #   AoA range.
            self.alpha0,
            self.alpha1,
            int(1+(self.alpha1-self.alpha0)/self.increment)
        )

        self.cases=[]
        for alpha in alpha_range:
            self.cases.append(Case( #   Creates case objects for range of alphas
                path=self.path,
                Xcg=self.Xcg,
                Ycg=self.Ycg,
                Zcg=self.Zcg,
                Ixx=self.Ixx,
                Iyy=self.Iyy,
                Izz=self.Izz,
                mass=self.mass,
                velocity=self.velocity,
                density=self.density,
                alpha=alpha,
                modes=self.modes,
                polars=self.polars
            ))
        
        return None

    def read_config(self,file:str)->None:
        """
        Reads aero config file.

        Arguments:
            file {string} -- Aero config file.
        """
        str_to_bool=lambda x:True if (x=="Y") else False

        with open(file,'r') as f:
            lines=f.readlines()
        lines=[line for line in lines if line[0]!="#" and line!="\n"]
        
        if lines[0].strip()!="AERO CONFIG":
            print(f"\u001b[31m[Error]\u001b[0m Wrong config file ({file}).")
            exit()

        lines=lines[1:]

        try:
            self.mass       = float(lines[0].split()[1])
            self.Xcg        = float(lines[1].split()[1])
            self.Ycg        = float(lines[2].split()[1])
            self.Zcg        = float(lines[3].split()[1])
            self.Ixx        = float(lines[4].split()[1])
            self.Iyy        = float(lines[5].split()[1])
            self.Izz        = float(lines[6].split()[1])
            self.velocity   = float(lines[7].split()[1])
            self.density    = float(lines[8].split()[1])
            self.alpha0     = float(lines[9].split()[1])
            self.alpha1     = float(lines[10].split()[1])
            self.increment  = float(lines[11].split()[1])
            self.threads    = int(lines[12].split()[1])
            self.polars     = str_to_bool(lines[13].split()[1])
            self.modes      = str_to_bool(lines[14].split()[1])
        except IndexError:
            print("Parameters must have a value assigned.")
            exit()
        

        return None

    def run(self,plane):
        """
        Writes cases and runs aero analyses.

        Arguments:
            plane {geometry.Plane} -- Plane object to run analysis on.
        """
        #   Write cases to file.
        with ThreadPoolExecutor(max_workers=self.threads) as pool:
            pool.map(self.create_cases,self.cases)

        plane.cases=self.cases

        #   Run aero analysis. Eigenmode and polar analysis both included.
        tasks=[(case,plane) for case in self.cases]
        with ThreadPoolExecutor(max_workers=self.threads) as pool:
            pool.map(self.analysis,tasks)

        # Both of these will be true because they're required.
        if self.modes==True:
            plane.modes=self.read_modes()
        if self.polars==True:
            plane.polars=self.read_aero()

        return None

    def create_cases(self,case):
        """Short function for multithreading sake."""
        case.write_aero_case()

        return None

    def analysis(self,tasks):
        """
        Writes command string and submits to AVL for polar and eigenmode analysis.

        Arguments:
            tasks {tuple[Case,Plane]} -- Case and plane to run analysis on.
        """
        case,plane=tasks

        cmd_str=f"load {plane.geom_file}\n"
        cmd_str+=f"case {case.case_file}\n"
        cmd_str+="oper\no\nv\n\nx\n"

        results_file=f"{self.path}/results/{plane.name}-{str(case.alpha)}deg"
        
        if case.modes==False and case.polars==False:
            raise ValueError("No analysis type defined.")

        if case.modes==True:
            case.modes_results_file=f"{results_file}.eig"

            cmd_str+="\nmode\nN\nW\n"
            cmd_str+=f"{case.modes_results_file}\n\n"
        if case.polars==True:
            case.polars_results_file=f"{results_file}.polars"

            cmd_str+="oper\nx\nst\n"
            cmd_str+=f"{case.polars_results_file}\n"

        avl_cmd(cmd_str,self.path)

        return None

    def read_aero(self):
        """
        Reads aero polar results files.

        Returns:
            polars_df {pd.DataFrame} -- Dataframe with polar and stab. derivative data for each alpha.
        """
        polars=[]
        for case in self.cases:
            with open(case.polars_results_file,'r') as file:
                lines=file.readlines()

                case.Cl=float(lines[23].split()[2])
                case.Cd=float(lines[24].split()[2])
                case.Clb=float(lines[38].split()[8])
                case.Clp=float(lines[46].split()[5])
                try:
                    case.spiral=float(lines[52].split()[6])
                except IndexError:
                    Cnb=float(lines[40].split()[8])
                    Clr=float(lines[46].split()[11])
                    Cnr=float(lines[48].split()[11])

                    try:
                        case.spiral=(case.Clb*Cnr)/(Clr*Cnb)
                    except ZeroDivisionError:
                        case.spiral=np.NaN
                
            polars.append((case.alpha,case.Cl,case.Cd,case.Clb,case.Clp,case.spiral))

        polars_df=pd.DataFrame(polars,columns=["Alpha (deg)","Cl","Cd","Clb","Clp","spiral"])

        return polars_df

    def read_modes(self):
        """
        Reads eigenmode results files.

        Returns:
            modes_df {pd.DataFrame} -- Dataframe with eigenmode data for each alpha.
        """
        modes=[]
        for case in self.cases:
            with open(case.modes_results_file,'r') as file:
                lines=file.readlines()

                #   AVL doesn't label which are which in results file and sometimes doesn't
                #   write them which is very annoying. Dutch roll and roll subsidence are the
                #   important ones and are consistently in the expected place in the file so
                #   everything else gets commented out ¯\_(ツ)_/¯
                try:
                    case.dutch=tuple(map(float,lines[3].split()[1:]))
                    #case.ndutch=tuple(map(float,lines[4].split()[1:]))
                    case.roll=tuple(map(float,lines[5].split()[1:]))
                    #case.short=tuple(map(float,lines[6].split()[1:]))
                    #case.nshort=tuple(map(float,lines[7].split()[1:]))
                    #case.lateral=tuple(map(float,lines[8].split()[1:]))
                    #case.phugoid=tuple(map(float,lines[9].split()[1:]))
                    #case.nphugoid=tuple(map(float,lines[10].split()[1:]))
                except IndexError as e:
                    print(f"Eigenmode analysis/read failed: Case {case.modes_results_file}")
                    print(f"\n{e}")
                    exit()

            modes.append((
                case.alpha,
                case.dutch,
                #case.ndutch,
                case.roll
                #case.short,
                #case.nshort,
                #case.lateral,
                #case.phugoid,
                #case.nphugoid
                ))
            
        modes_df=pd.DataFrame(modes,columns=[
            "alpha",
            "dutch",
            #"-dutch",
            "roll"
            #"short",
            #"-short",
            #"lateral",
            #"phugoid",
            #"-phugoid"
            ])

        return modes_df

if __name__=="__main__":
    plane=Plane(geom_file='example_plane.avl')

    aero=Aero("aero.config")
    aero.run(plane)

    print(plane.polars)
    print(plane.modes)



from setuptools import setup, find_packages
from pathlib import Path

VERSION = '1.0.2'
DESCRIPTION = "Automation tools for vortex lattice method, AVL"
LONG_DESCRIPTION = (Path(__file__).parent / "README.md").read_text()

# setup
setup(
    name="avlautomation",
    version=VERSION,
    author="jj-foster",
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url="https://github.com/Team-Peryton/AVL-automation",
    packages=find_packages(),
    install_requires=["numpy","scipy","pandas","matplotlib","tqdm"],
    python_requires='>=3.9'
)
